import { isPromise } from '../jsutils/isPromise.mjs';
import { pathToArray } from '../jsutils/Path.mjs';
import { promiseWithResolvers } from '../jsutils/promiseWithResolvers.mjs';
export function buildIncrementalResponse(
  result,
  errors,
  futures,
  cancellableStreams,
) {
  const incrementalPublisher = new IncrementalPublisher(cancellableStreams);
  return incrementalPublisher.buildResponse(result, errors, futures);
}
/**
 * This class is used to publish incremental results to the client, enabling semi-concurrent
 * execution while preserving result order.
 *
 * @internal
 */
class IncrementalPublisher {
  constructor(cancellableStreams) {
    this._cancellableStreams = cancellableStreams;
    this._nextId = 0;
    this._pending = new Set();
    this._completedResultQueue = [];
    this._newPending = new Set();
    this._incremental = [];
    this._completed = [];
    this._reset();
  }
  buildResponse(data, errors, futures) {
    this._addFutures(futures);
    this._pruneEmpty();
    const pending = this._pendingSourcesToResults();
    const initialResult =
      errors.length === 0
        ? { data, pending, hasNext: true }
        : { errors, data, pending, hasNext: true };
    return {
      initialResult,
      subsequentResults: this._subscribe(),
    };
  }
  _addFutures(futures) {
    for (const future of futures) {
      if (isDeferredGroupedFieldSetRecord(future)) {
        for (const deferredFragmentRecord of future.deferredFragmentRecords) {
          this._addDeferredFragmentRecord(deferredFragmentRecord);
        }
        const result = future.result;
        if (isPromise(result)) {
          // eslint-disable-next-line @typescript-eslint/no-floating-promises
          result.then((resolved) => {
            this._enqueueCompletedDeferredGroupedFieldSet(resolved);
          });
        } else {
          this._enqueueCompletedDeferredGroupedFieldSet(result);
        }
        continue;
      }
      const streamRecord = future.streamRecord;
      if (streamRecord.id === undefined) {
        this._newPending.add(streamRecord);
      }
      const result = future.getResult();
      if (isPromise(result)) {
        // eslint-disable-next-line @typescript-eslint/no-floating-promises
        result.then((resolved) => {
          this._enqueueCompletedStreamItems(resolved);
        });
      } else {
        this._enqueueCompletedStreamItems(result);
      }
    }
  }
  _addDeferredFragmentRecord(deferredFragmentRecord) {
    const parent = deferredFragmentRecord.parent;
    if (parent === undefined) {
      if (deferredFragmentRecord.id !== undefined) {
        return;
      }
      this._newPending.add(deferredFragmentRecord);
      return;
    }
    if (parent.children.has(deferredFragmentRecord)) {
      return;
    }
    parent.children.add(deferredFragmentRecord);
    this._addDeferredFragmentRecord(parent);
  }
  _pruneEmpty() {
    const maybeEmptyNewPending = this._newPending;
    this._newPending = new Set();
    for (const node of maybeEmptyNewPending) {
      if (isDeferredFragmentRecord(node)) {
        if (node.deferredGroupedFieldSetRecords.length > 0) {
          this._newPending.add(node);
          continue;
        }
        for (const child of node.children) {
          this._addNonEmptyNewPending(child);
        }
      } else {
        this._newPending.add(node);
      }
    }
  }
  _addNonEmptyNewPending(deferredFragmentRecord) {
    if (deferredFragmentRecord.deferredGroupedFieldSetRecords.length > 0) {
      this._newPending.add(deferredFragmentRecord);
      return;
    }
    /* c8 ignore next 5 */
    // TODO: add test case for this, if when skipping an empty deferred fragment, the empty fragment has nested children.
    for (const child of deferredFragmentRecord.children) {
      this._addNonEmptyNewPending(child);
    }
  }
  _enqueueCompletedDeferredGroupedFieldSet(result) {
    let hasPendingParent = false;
    for (const deferredFragmentRecord of result.deferredFragmentRecords) {
      if (deferredFragmentRecord.id !== undefined) {
        hasPendingParent = true;
      }
      deferredFragmentRecord.results.push(result);
    }
    if (hasPendingParent) {
      this._completedResultQueue.push(result);
      this._trigger();
    }
  }
  _enqueueCompletedStreamItems(result) {
    this._completedResultQueue.push(result);
    this._trigger();
  }
  _pendingSourcesToResults() {
    const pendingResults = [];
    for (const pendingSource of this._newPending) {
      const id = String(this._getNextId());
      this._pending.add(pendingSource);
      pendingSource.id = id;
      const pendingResult = {
        id,
        path: pathToArray(pendingSource.path),
      };
      if (pendingSource.label !== undefined) {
        pendingResult.label = pendingSource.label;
      }
      pendingResults.push(pendingResult);
    }
    this._newPending.clear();
    return pendingResults;
  }
  _getNextId() {
    return String(this._nextId++);
  }
  _subscribe() {
    let isDone = false;
    const _next = async () => {
      while (!isDone) {
        let pending = [];
        let completedResult;
        while (
          (completedResult = this._completedResultQueue.shift()) !== undefined
        ) {
          if (isDeferredGroupedFieldSetResult(completedResult)) {
            this._handleCompletedDeferredGroupedFieldSet(completedResult);
          } else {
            this._handleCompletedStreamItems(completedResult);
          }
          pending = [...pending, ...this._pendingSourcesToResults()];
        }
        if (this._incremental.length > 0 || this._completed.length > 0) {
          const hasNext = this._pending.size > 0;
          if (!hasNext) {
            isDone = true;
          }
          const subsequentIncrementalExecutionResult = { hasNext };
          if (pending.length > 0) {
            subsequentIncrementalExecutionResult.pending = pending;
          }
          if (this._incremental.length > 0) {
            subsequentIncrementalExecutionResult.incremental =
              this._incremental;
          }
          if (this._completed.length > 0) {
            subsequentIncrementalExecutionResult.completed = this._completed;
          }
          this._incremental = [];
          this._completed = [];
          return { value: subsequentIncrementalExecutionResult, done: false };
        }
        // eslint-disable-next-line no-await-in-loop
        await this._signalled;
      }
      await returnStreamIterators().catch(() => {
        // ignore errors
      });
      return { value: undefined, done: true };
    };
    const returnStreamIterators = async () => {
      const promises = [];
      for (const streamRecord of this._cancellableStreams) {
        if (streamRecord.earlyReturn !== undefined) {
          promises.push(streamRecord.earlyReturn());
        }
      }
      await Promise.all(promises);
    };
    const _return = async () => {
      isDone = true;
      await returnStreamIterators();
      return { value: undefined, done: true };
    };
    const _throw = async (error) => {
      isDone = true;
      await returnStreamIterators();
      return Promise.reject(error);
    };
    return {
      [Symbol.asyncIterator]() {
        return this;
      },
      next: _next,
      return: _return,
      throw: _throw,
    };
  }
  _trigger() {
    this._resolve();
    this._reset();
  }
  _reset() {
    // promiseWithResolvers uses void only as a generic type parameter
    // see: https://typescript-eslint.io/rules/no-invalid-void-type/
    // eslint-disable-next-line @typescript-eslint/no-invalid-void-type
    const { promise: signalled, resolve } = promiseWithResolvers();
    this._resolve = resolve;
    this._signalled = signalled;
  }
  _handleCompletedDeferredGroupedFieldSet(result) {
    if (!isReconcilableDeferredGroupedFieldSetResult(result)) {
      for (const deferredFragmentRecord of result.deferredFragmentRecords) {
        const id = deferredFragmentRecord.id;
        if (id !== undefined) {
          this._completed.push({ id, errors: result.errors });
          this._pending.delete(deferredFragmentRecord);
        }
      }
      return;
    }
    for (const deferredFragmentRecord of result.deferredFragmentRecords) {
      deferredFragmentRecord.reconcilableResults.push(result);
    }
    if (result.futures) {
      this._addFutures(result.futures);
    }
    for (const deferredFragmentRecord of result.deferredFragmentRecords) {
      const id = deferredFragmentRecord.id;
      // TODO: add test case for this.
      // Presumably, this can occur if an error causes a fragment to be completed early,
      // while an asynchronous deferred grouped field set result is enqueued.
      /* c8 ignore next 3 */
      if (id === undefined) {
        continue;
      }
      const fragmentResults = deferredFragmentRecord.reconcilableResults;
      if (
        deferredFragmentRecord.deferredGroupedFieldSetRecords.length !==
        fragmentResults.length
      ) {
        continue;
      }
      for (const fragmentResult of fragmentResults) {
        if (fragmentResult.sent) {
          continue;
        }
        fragmentResult.sent = true;
        const { bestId, subPath } = this._getBestIdAndSubPath(
          id,
          deferredFragmentRecord,
          fragmentResult,
        );
        const incrementalEntry = {
          data: fragmentResult.data,
          id: bestId,
        };
        if (result.errors.length > 0) {
          incrementalEntry.errors = fragmentResult.errors;
        }
        if (subPath !== undefined) {
          incrementalEntry.subPath = subPath;
        }
        this._incremental.push(incrementalEntry);
      }
      this._completed.push({ id });
      this._pending.delete(deferredFragmentRecord);
      for (const child of deferredFragmentRecord.children) {
        this._newPending.add(child);
        for (const childResult of child.results) {
          if (!isPromise(childResult)) {
            this._completedResultQueue.push(childResult);
          }
        }
      }
    }
    this._pruneEmpty();
  }
  _handleCompletedStreamItems(result) {
    const streamRecord = result.streamRecord;
    const id = streamRecord.id;
    // TODO: Consider adding invariant or non-null assertion, as this should never happen. Since the stream is converted into a linked list
    // for ordering purposes, if an entry errors, additional entries will not be processed.
    /* c8 ignore next 3 */
    if (id === undefined) {
      return;
    }
    if (result.items === undefined) {
      this._completed.push({ id });
      this._pending.delete(streamRecord);
      this._cancellableStreams.delete(streamRecord);
    } else if (result.items === null) {
      this._completed.push({
        id,
        errors: result.errors,
      });
      this._pending.delete(streamRecord);
      this._cancellableStreams.delete(streamRecord);
      streamRecord.earlyReturn?.().catch(() => {
        /* c8 ignore next 1 */
        // ignore error
      });
    } else {
      const incrementalEntry = {
        id,
        items: result.items, // FIX!
      };
      if (result.errors !== undefined && result.errors.length > 0) {
        incrementalEntry.errors = result.errors;
      }
      this._incremental.push(incrementalEntry);
      if (result.futures) {
        this._addFutures(result.futures);
        this._pruneEmpty();
      }
    }
  }
  _getBestIdAndSubPath(
    initialId,
    initialDeferredFragmentRecord,
    deferredGroupedFieldSetResult,
  ) {
    let maxLength = pathToArray(initialDeferredFragmentRecord.path).length;
    let bestId = initialId;
    for (const deferredFragmentRecord of deferredGroupedFieldSetResult.deferredFragmentRecords) {
      if (deferredFragmentRecord === initialDeferredFragmentRecord) {
        continue;
      }
      const id = deferredFragmentRecord.id;
      // TODO: add test case for when an fragment has not been released, but might be processed for the shortest path.
      /* c8 ignore next 3 */
      if (id === undefined) {
        continue;
      }
      const fragmentPath = pathToArray(deferredFragmentRecord.path);
      const length = fragmentPath.length;
      if (length > maxLength) {
        maxLength = length;
        bestId = id;
      }
    }
    const subPath = deferredGroupedFieldSetResult.path.slice(maxLength);
    return {
      bestId,
      subPath: subPath.length > 0 ? subPath : undefined,
    };
  }
}
export function isDeferredFragmentRecord(subsequentResultRecord) {
  return subsequentResultRecord instanceof DeferredFragmentRecord;
}
export function isDeferredGroupedFieldSetRecord(future) {
  return future instanceof DeferredGroupedFieldSetRecord;
}
export function isDeferredGroupedFieldSetResult(subsequentResult) {
  return 'deferredFragmentRecords' in subsequentResult;
}
export function isReconcilableDeferredGroupedFieldSetResult(
  deferredGroupedFieldSetResult,
) {
  return deferredGroupedFieldSetResult.data !== null;
}
/** @internal */
export class DeferredGroupedFieldSetRecord {
  constructor(opts) {
    const { path, deferUsageSet, deferredFragmentRecords, executor } = opts;
    this.path = path;
    this.deferredFragmentRecords = deferredFragmentRecords;
    const incrementalContext = {
      deferUsageSet,
      errors: [],
      errorPaths: new Set(),
      futures: [],
    };
    for (const deferredFragmentRecord of this.deferredFragmentRecords) {
      deferredFragmentRecord.deferredGroupedFieldSetRecords.push(this);
    }
    this.result = this.deferredFragmentRecords.some(
      (deferredFragmentRecord) => deferredFragmentRecord.id !== undefined,
    )
      ? executor(incrementalContext)
      : Promise.resolve().then(() => executor(incrementalContext));
  }
}
/** @internal */
export class DeferredFragmentRecord {
  constructor(opts) {
    this.path = opts.path;
    this.label = opts.label;
    this.deferredGroupedFieldSetRecords = [];
    this.results = [];
    this.reconcilableResults = [];
    this.parent = opts.parent;
    this.children = new Set();
  }
}
/** @internal */
export class StreamRecord {
  constructor(opts) {
    const { label, path, earlyReturn } = opts;
    this.label = label;
    this.path = path;
    this.earlyReturn = earlyReturn;
  }
}
/** @internal */
export class StreamItemsRecord {
  constructor(opts) {
    const { streamRecord, executor } = opts;
    this.streamRecord = streamRecord;
    const incrementalContext = {
      deferUsageSet: undefined,
      errors: [],
      errorPaths: new Set(),
      futures: [],
    };
    this._result = executor(incrementalContext);
  }
  getResult() {
    if (isPromise(this._result)) {
      return this._result.then((resolved) =>
        this._prependNextStreamItems(resolved),
      );
    }
    return this._prependNextStreamItems(this._result);
  }
  _prependNextStreamItems(result) {
    return this.nextStreamItems === undefined
      ? result
      : {
          ...result,
          futures: [this.nextStreamItems, ...(result.futures ?? [])],
        };
  }
}
